from datetime import datetime, timezone
from typing import Any, Dict, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator, model_validator

from whiskerrag_types.model.timeStampedModel import TimeStampedModel


class SpaceCreate(BaseModel):
    """
    SpaceCreate model for creating space resources.
    Attributes:
        space_name (str): Space name, example: petercat bot group.
        description (str): descrition of the space resource.
        metadata (Dict[str, Any]): metadata of the space resource.such as embedding model name
            and other parameters.
    """

    space_name: str = Field(
        ..., max_length=64, description="name of the space resource"
    )
    space_id: Optional[str] = Field(
        default=None,
        description="space id, e.g. petercat/bot-group",
        pattern=r"^[A-Za-z0-9._@/-]{1,255}$",
        max_length=255,
    )
    description: str = Field(..., max_length=255, description="descrition of the space")
    metadata: Dict[str, Any] = Field(
        default={},
        description="metadata of the space resource",
    )

    @field_validator("space_id")
    @classmethod
    def check_forbidden_sequences(cls, v: str) -> str:
        if v is None:
            return v
        # 禁止连续 ..
        if ".." in v:
            raise ValueError("space_id cannot contain consecutive dots '..'")
        # 禁止连续 //
        if "//" in v:
            raise ValueError("space_id cannot contain consecutive slashes '//'")
        return v


class Space(SpaceCreate, TimeStampedModel):
    space_id: str = Field(default_factory=lambda: str(uuid4()), description="space id")
    tenant_id: str = Field(..., description="tenant id")

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    def update(self, **kwargs: Dict[str, Any]) -> "Space":
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.updated_at = datetime.now(timezone.utc)
        return self

    @model_validator(mode="before")
    def pre_process_data(cls, data: dict) -> dict:
        for field, value in data.items():
            if isinstance(value, UUID):
                data[field] = str(value)
        if isinstance(data, dict) and not data.get("space_id"):
            data["space_id"] = str(uuid4())
        return data


class SpaceResponse(Space):
    """
    SpaceResponse model class that extends Space.
    Attributes:
         (str): Space ID.
        total_size Optional[int]: size of the all kowledge in this space.
        knowledge_size Optional[int]: count of the knowledge in this space.
    """

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)

    storage_size: Optional[int] = Field(
        default=0, description="size of the all kowledge in this space"
    )
    knowledge_count: Optional[int] = Field(
        default=0, description="count of the knowledge in this space"
    )

    def update(self, **kwargs: Dict[str, Any]) -> "SpaceResponse":
        for key, value in kwargs.items():
            setattr(self, key, value)
        return self



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar, Union

from whiskerrag_types.model.knowledge import Knowledge
from whiskerrag_types.model.multi_modal import Blob, Image, Text

T = TypeVar("T", Image, Text, Blob, Union[Image, Text, Blob])


class BaseLoader(ABC, Generic[T]):
    def __init__(self, knowledge: Knowledge):
        self.knowledge = knowledge

    @abstractmethod
    async def load(self) -> List[T]:
        """
        Load the knowledge into a list of items.
        This method should be implemented by subclasses to define how the knowledge is loaded.
        """
        pass

    @abstractmethod
    async def decompose(self) -> List[Knowledge]:
        """
        Decompose the knowledge into smaller parts.
        This method should be implemented by subclasses to define how the knowledge is decomposed.
        """
        pass

    @abstractmethod
    async def on_load_finished(self) -> None:
        """
        Lifecycle method called when the loading task is finished.
        Subclasses can implement this to perform any cleanup or post-processing.
        """
        pass



from .agent import ChatCompletionMessageParam, KnowledgeScope, ProResearchRequest
from .api_key import APIKey
from .chunk import Chunk
from .converter import GenericConverter
from .knowledge import (
    EmbeddingModelEnum,
    Knowledge,
    KnowledgeSourceEnum,
    KnowledgeTypeEnum,
)
from .knowledge_create import (
    KNOWLEDGE_CREATE_2_KNOWLEDGE_STRATEGY_MAP,
    GithubRepoCreate,
    ImageCreate,
    JSONCreate,
    KnowledgeCreateUnion,
    MarkdownCreate,
    PDFCreate,
    QACreate,
    TextCreate,
)
from .knowledge_source import (
    GithubFileSourceConfig,
    GithubRepoSourceConfig,
    OpenUrlSourceConfig,
    S3SourceConfig,
    TextSourceConfig,
    YuqueSourceConfig,
)
from .language import LanguageEnum
from .page import (
    PageParams,
    PageQueryParams,
    PageResponse,
    QueryParams,
    StatusStatisticsPageResponse,
)
from .permission import Action, Permission, Resource
from .retrieval import (
    RetrievalByKnowledgeRequest,
    RetrievalBySpaceRequest,
    RetrievalChunk,
    RetrievalRequest,
)
from .rule import GlobalRule, Rule, SpaceRule
from .space import Space, SpaceCreate, SpaceResponse
from .splitter import (
    BaseCharSplitConfig,
    BaseCodeSplitConfig,
    GeaGraphSplitConfig,
    GithubRepoParseConfig,
    ImageSplitConfig,
    JSONSplitConfig,
    KnowledgeSplitConfig,
    MarkdownSplitConfig,
    PDFSplitConfig,
    TextSplitConfig,
    YuqueSplitConfig,
)
from .tag import Tag, TagCreate
from .tagging import Tagging, TaggingCreate, TagObjectType
from .task import Task, TaskRestartRequest, TaskStatus
from .tenant import Tenant
from .wiki import Wiki

__all__ = [
    "APIKey",
    "Action",
    "Resource",
    "Permission",
    "Chunk",
    "ChatCompletionMessageParam",
    "Rule",
    "GlobalRule",
    "SpaceRule",
    "KnowledgeSourceEnum",
    "KnowledgeTypeEnum",
    "EmbeddingModelEnum",
    "KnowledgeSplitConfig",
    "TextCreate",
    "ImageCreate",
    "JSONCreate",
    "MarkdownCreate",
    "OpenUrlSourceConfig",
    "PDFCreate",
    "GithubRepoCreate",
    "QACreate",
    "KnowledgeCreateUnion",
    "GithubRepoSourceConfig",
    "GithubFileSourceConfig",
    "S3SourceConfig",
    "TextSourceConfig",
    "YuqueSourceConfig",
    "Knowledge",
    "Space",
    "SpaceCreate",
    "SpaceResponse",
    "LanguageEnum",
    "PageQueryParams",
    "PageParams",
    "QueryParams",
    "PageResponse",
    "StatusStatisticsPageResponse",
    "RetrievalBySpaceRequest",
    "RetrievalByKnowledgeRequest",
    "RetrievalChunk",
    "RetrievalRequest",
    "Task",
    "TaskStatus",
    "TaskRestartRequest",
    "Tenant",
    "GenericConverter",
    "BaseCharSplitConfig",
    "JSONSplitConfig",
    "MarkdownSplitConfig",
    "PDFSplitConfig",
    "TextSplitConfig",
    "GeaGraphSplitConfig",
    "YuqueSplitConfig",
    "ImageSplitConfig",
    "BaseCodeSplitConfig",
    "GithubRepoParseConfig",
    "Wiki",
    "KNOWLEDGE_CREATE_2_KNOWLEDGE_STRATEGY_MAP",
    "KnowledgeScope",
    "ProResearchRequest",
    "Tag",
    "TagCreate",
    "Tagging",
    "TagObjectType",
    "TaggingCreate",
]



"""
异步转同步的通用工具模块

提供了在不同环境下（有/无事件循环）安全执行异步代码的工具函数。
"""

import asyncio
import functools
from typing import Any, Awaitable, Callable, Coroutine, TypeVar, cast

T = TypeVar("T")


def run_async_safe(
    coro_func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
) -> T:
    """
    安全地运行异步函数，自动处理事件循环检测

    这个函数会：
    1. 检测当前是否已经在事件循环中
    2. 如果在事件循环中，使用 asyncio.create_task 在当前循环中执行
    3. 如果不在事件循环中，直接使用 asyncio.run

    Args:
        coro_func: 异步函数
        *args: 传递给异步函数的位置参数
        **kwargs: 传递给异步函数的关键字参数

    Returns:
        异步函数的返回值

    Raises:
        Exception: 异步函数执行过程中的任何异常

    Example:
        >>> async def async_task(x: int) -> int:
        ...     await asyncio.sleep(0.1)
        ...     return x * 2
        >>>
        >>> result = run_async_safe(async_task, 5)
        >>> print(result)  # 10
    """
    try:
        # 检查是否已经在事件循环中
        asyncio.get_running_loop()
        # 如果在事件循环中，需要在新线程中运行
        import concurrent.futures

        def run_in_thread() -> T:
            # 将 Awaitable 转换为 Coroutine 以满足 asyncio.run 的类型要求
            coro = coro_func(*args, **kwargs)
            return asyncio.run(cast(Coroutine[Any, Any, T], coro))

        with concurrent.futures.ThreadPoolExecutor() as executor:
            future = executor.submit(run_in_thread)
            return future.result()

    except RuntimeError:
        # 没有运行中的事件循环，可以直接使用 asyncio.run
        coro = coro_func(*args, **kwargs)
        return asyncio.run(cast(Coroutine[Any, Any, T], coro))


def async_to_sync(coro_func: Callable[..., Awaitable[T]]) -> Callable[..., T]:
    """
    装饰器：将异步函数转换为同步函数

    这个装饰器会自动处理事件循环检测，让异步函数可以在同步环境中安全调用。

    Args:
        coro_func: 要转换的异步函数

    Returns:
        转换后的同步函数

    Example:
        >>> @async_to_sync
        ... async def async_task(x: int) -> int:
        ...     await asyncio.sleep(0.1)
        ...     return x * 2
        >>>
        >>> result = async_task(5)  # 现在可以同步调用
        >>> print(result)  # 10
    """

    @functools.wraps(coro_func)
    def wrapper(*args: Any, **kwargs: Any) -> T:
        return run_async_safe(coro_func, *args, **kwargs)

    return wrapper


class AsyncToSyncMixin:
    """
    混入类：为类提供异步转同步的能力

    继承这个混入类后，可以使用 _run_async_safe 方法来安全执行异步代码。

    Example:
        >>> class MyClass(AsyncToSyncMixin):
        ...     async def async_method(self, x: int) -> int:
        ...         await asyncio.sleep(0.1)
        ...         return x * 2
        ...
        ...     def sync_method(self, x: int) -> int:
        ...         return self._run_async_safe(self.async_method, x)
        >>>
        >>> obj = MyClass()
        >>> result = obj.sync_method(5)
        >>> print(result)  # 10
    """

    def _run_async_safe(
        self, coro_func: Callable[..., Awaitable[T]], *args: Any, **kwargs: Any
    ) -> T:
        """
        安全地运行异步方法

        Args:
            coro_func: 异步方法
            *args: 传递给异步方法的位置参数
            **kwargs: 传递给异步方法的关键字参数

        Returns:
            异步方法的返回值
        """
        return run_async_safe(coro_func, *args, **kwargs)



from typing import List, Union

from langchain_text_splitters import RecursiveCharacterTextSplitter

from whiskerrag_types.interface.parser_interface import BaseParser, ParseResult
from whiskerrag_types.model.knowledge import Knowledge
from whiskerrag_types.model.multi_modal import Image, Text
from whiskerrag_types.model.splitter import BaseCodeSplitConfig
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.PARSER, "base_code")
class CodeParser(BaseParser[Text]):

    def _calculate_line_position(
        self, content: str, chunk: str, char_start: int
    ) -> dict:
        """
        Calculate line-based position information for a code chunk

        Args:
            content: Original full content
            chunk: The code chunk
            char_start: Character start position in the original content

        Returns:
            dict: Position information including line numbers
        """
        # Split content into lines for position calculation
        lines = content.split("\n")

        # Calculate start line
        current_char = 0
        start_line = 1
        start_column = 1

        for line_num, line in enumerate(lines, 1):
            line_length = len(line) + 1  # +1 for newline character
            if current_char + line_length > char_start:
                start_line = line_num
                start_column = char_start - current_char + 1
                break
            current_char += line_length

        # Calculate end line
        chunk_lines = chunk.split("\n")
        end_line = start_line + len(chunk_lines) - 1

        # Calculate end column
        if len(chunk_lines) == 1:
            end_column = start_column + len(chunk_lines[0]) - 1
        else:
            end_column = len(chunk_lines[-1])

        return {
            "start_line": start_line,
            "end_line": end_line,
            "start_column": start_column,
            "end_column": end_column,
            "total_lines": len(chunk_lines),
        }

    async def parse(
        self,
        knowledge: Knowledge,
        content: Text,
    ) -> ParseResult:
        split_config = knowledge.split_config
        if not isinstance(split_config, BaseCodeSplitConfig):
            raise TypeError("knowledge.split_config must be of type CodeSplitConfig")
        splitter = RecursiveCharacterTextSplitter.from_language(
            split_config.language,
            chunk_size=split_config.chunk_size,
            chunk_overlap=split_config.chunk_overlap,
        )
        split_docs = splitter.split_text(
            content.content,
        )

        # Create Text objects with proper metadata inheritance and position info
        results: List[Union[Text, Image]] = []
        current_position = 0

        for chunk_index, doc in enumerate(split_docs):
            # Start with knowledge.metadata as base
            combined_metadata = {**knowledge.metadata}

            # Add Text.metadata from content (loader/previous parser stage)
            if content.metadata:
                combined_metadata.update(content.metadata)

            # Find the actual start position of this chunk in the original content
            doc_start = content.content.find(doc, current_position)
            if doc_start == -1:
                # Fallback if exact match not found
                doc_start = current_position

            doc_end = doc_start + len(doc)

            # Calculate line-based position information
            line_position = self._calculate_line_position(
                content.content, doc, doc_start
            )
            # Add processing information from current parser stage
            parser_metadata = {
                "chunk_index": chunk_index,
                "parser_type": "base_code",
                "position": line_position,
            }

            combined_metadata.update(parser_metadata)

            results.append(Text(content=doc, metadata=combined_metadata))
            current_position = doc_end

        return results

    async def batch_parse(
        self,
        knowledge: Knowledge,
        content_list: List[Text],
    ) -> List[ParseResult]:
        return [await self.parse(knowledge, content) for content in content_list]



from datetime import datetime, timezone
from typing import Any, Dict, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator, model_validator

from whiskerrag_types.model.timeStampedModel import TimeStampedModel


class ArtifactIndexCreate(BaseModel):
    """
    创建 whisker_artifact_index 条目的入参模型
    """

    ecosystem: str = Field(
        ...,
        max_length=32,
        description="制品来源生态系统（pypi / npm / maven / go / php）",
    )
    name: str = Field(
        ...,
        max_length=255,
        description="制品名（构建产物名，如 requests / @company/sdk）",
    )
    version: Optional[str] = Field(
        default=None, max_length=64, description="版本号（可为空）"
    )
    space_id: str = Field(
        ...,
        max_length=255,
        pattern=r"^[A-Za-z0-9._@/-]{1,255}$",
        description="关联的 whisker_space.space_id",
    )
    extra: Dict[str, Any] = Field(
        default={}, description="额外元数据信息，扩展用，如构建参数、标签、扫描信息等"
    )

    @field_validator("space_id")
    @classmethod
    def check_forbidden_sequences(cls, v: str) -> str:
        if ".." in v:
            raise ValueError("space_id cannot contain consecutive dots '..'")
        if "//" in v:
            raise ValueError("space_id cannot contain consecutive slashes '//'")
        return v

    @field_validator("ecosystem")
    @classmethod
    def normalize_ecosystem(cls, v: str) -> str:
        if isinstance(v, str):
            return v.strip().lower()
        return v

    @field_validator("name")
    @classmethod
    def normalize_name(cls, v: str) -> str:
        return v.strip() if isinstance(v, str) else v


class ArtifactIndex(ArtifactIndexCreate, TimeStampedModel):
    """
    whisker_artifact_index 模型
    """

    artifact_id: str = Field(
        default_factory=lambda: str(uuid4()), description="制品索引表主键（UUID字符串）"
    )

    def update(self, **kwargs: Any) -> "ArtifactIndex":
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.updated_at = datetime.now(timezone.utc)
        return self

    @model_validator(mode="before")
    def preprocess(cls, data: dict) -> dict:
        for field, value in list(data.items()):
            if isinstance(value, UUID):
                data[field] = str(value)
        return data



from enum import Enum
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from deprecated import deprecated
from pydantic import BaseModel, Field, model_validator

T = TypeVar("T")


class Operator(str, Enum):
    AND = "and"
    OR = "or"


class Condition(BaseModel):
    field: str
    operator: str  # eq, neq, gt, gte, lt, lte, like, ilike
    value: Any


class FilterGroup(BaseModel):
    operator: Operator
    conditions: List[Union[Condition, "FilterGroup"]]


class QueryParams(BaseModel, Generic[T]):
    order_by: Optional[str] = Field(default=None, description="order by field")
    order_direction: Optional[str] = Field(default="asc", description="asc or desc")
    eq_conditions: Optional[Dict[str, Any]] = Field(
        default=None,
        description="list of equality conditions, each as a dict with key and value",
    )
    advanced_filter: Optional[FilterGroup] = Field(
        default=None,
        description="advanced filter with nested conditions",
    )

    def _validate_fields_against_model(self, fields: set[str]) -> set[str]:
        """validate fields against model"""
        args = self.__class__.__pydantic_generic_metadata__["args"]
        if not args:
            return set()

        model_type = args[0]
        if isinstance(model_type, TypeVar):
            return set()

        # Get all valid field names including aliases
        valid_fields = set()
        for field_name, field_info in model_type.model_fields.items():
            valid_fields.add(field_name)
            # Add alias if it exists
            if field_info.alias:
                valid_fields.add(field_info.alias)

        return fields - valid_fields

    def _validate_filter_group(self, filter_group: FilterGroup) -> set[str]:
        """recursively validate all fields in FilterGroup"""
        invalid_fields = set()

        for condition in filter_group.conditions:
            if isinstance(condition, Condition):
                invalid_fields.add(condition.field)
            elif isinstance(condition, FilterGroup):
                invalid_fields.update(self._validate_filter_group(condition))

        return invalid_fields

    @model_validator(mode="after")
    def validate_conditions(self) -> "QueryParams[T]":
        invalid_fields = set()

        # validate eq_conditions
        if self.eq_conditions:
            invalid_fields.update(
                self._validate_fields_against_model(set(self.eq_conditions.keys()))
            )

        # validate advanced_filter
        if self.advanced_filter:
            filter_fields = self._validate_filter_group(self.advanced_filter)
            invalid_fields.update(self._validate_fields_against_model(filter_fields))

        # if invalid fields, raise exception
        if invalid_fields:
            raise ValueError(f"Invalid fields found: {invalid_fields}")

        return self


class PageQueryParams(QueryParams[T], Generic[T]):
    page: int = Field(default=1, ge=1, description="page number")
    page_size: int = Field(default=10, ge=1, le=1000, description="page size")

    @property
    def offset(self) -> int:
        return (self.page - 1) * self.page_size

    @property
    def limit(self) -> int:
        return self.page_size


@deprecated(reason="Use PageQueryParams instead")
class PageParams(PageQueryParams[T], Generic[T]):
    pass


class PageResponse(BaseModel, Generic[T]):
    items: List[T]
    total: int
    page: int
    page_size: int
    total_pages: int


class StatusStatisticsPageResponse(PageResponse, Generic[T]):
    """
    TaskStatus
    """

    success: int = 0
    failed: int = 0
    cancelled: int = 0
    pending: int = 0
    running: int = 0
    pending_retry: int = 0
    deleted: int = 0



import logging
from abc import ABC, abstractmethod
from typing import Any, AsyncIterator, List, Optional, TypeVar, Union

from pydantic import BaseModel

from whiskerrag_types.model import (
    APIKey,
    Knowledge,
    PageQueryParams,
    PageResponse,
    Space,
    Task,
    TaskStatus,
    Tenant,
)
from whiskerrag_types.model.artifact_index import ArtifactIndex, ArtifactIndexCreate
from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.retrieval import (
    RetrievalByKnowledgeRequest,
    RetrievalBySpaceRequest,
    RetrievalChunk,
    RetrievalRequest,
)
from whiskerrag_types.model.tag import Tag, TagCreate
from whiskerrag_types.model.tagging import Tagging, TaggingCreate

from .settings_interface import SettingsInterface

T = TypeVar("T", bound=BaseModel)


class DBPluginInterface(ABC):
    settings: SettingsInterface

    def __init__(self, settings: SettingsInterface) -> None:
        self.settings = settings
        self.logger = logging.getLogger("whisker")
        self._initialized: bool = False

    async def ensure_initialized(self) -> None:
        if not self._initialized:
            try:
                self.logger.info("DBEngine plugin is initializing...")
                await self.init()
                self._initialized = True
                self.logger.info("DBEngine plugin is initialized")
            except Exception as e:
                self.logger.error(f"DBEngine plugin init error: {e}")
                raise

    @property
    def is_initialized(self) -> bool:
        return self._initialized

    @abstractmethod
    async def init(self) -> None:
        pass

    @abstractmethod
    async def cleanup(self) -> None:
        pass

    @abstractmethod
    def get_db_client(self) -> Any:
        pass

    # =================== knowledge ===================
    @abstractmethod
    async def save_knowledge_list(
        self, knowledge_list: List[Knowledge]
    ) -> List[Knowledge]:
        pass

    @abstractmethod
    async def get_knowledge_list(
        self, tenant_id: str, page_params: PageQueryParams[Knowledge]
    ) -> PageResponse[Knowledge]:
        pass

    @abstractmethod
    async def get_knowledge(self, tenant_id: str, knowledge_id: str) -> Knowledge:
        pass

    @abstractmethod
    async def update_knowledge(self, knowledge: Knowledge) -> None:
        pass

    @abstractmethod
    async def delete_knowledge(
        self, tenant_id: str, knowledge_id_list: List[str], cascade: bool = False
    ) -> None:
        pass

    @abstractmethod
    async def batch_update_knowledge_retrieval_count(
        self, knowledge_id_list: dict[str, int]
    ) -> None:
        pass

    @abstractmethod
    async def update_knowledge_enabled_status(
        self, tenant_id: str, knowledge_id: str, enabled: bool
    ) -> None:
        pass

    # =================== Space ===================
    @abstractmethod
    async def save_space(self, space: Space) -> Space:
        pass

    @abstractmethod
    async def update_space(self, space: Space) -> Space:
        pass

    @abstractmethod
    async def get_space_list(
        self, tenant_id: str, page_params: PageQueryParams[Space]
    ) -> PageResponse[Space]:
        pass

    @abstractmethod
    async def get_space(self, tenant_id: str, space_id: str) -> Space:
        pass

    @abstractmethod
    async def delete_space(
        self, tenant_id: str, space_id: str
    ) -> Union[List[Space], None]:
        pass

    # =================== chunk ===================
    @abstractmethod
    async def save_chunk_list(self, chunks: List[Chunk]) -> List[Chunk]:
        pass

    @abstractmethod
    async def update_chunk_list(self, chunks: List[Chunk]) -> List[Chunk]:
        pass

    @abstractmethod
    async def get_chunk_list(
        self, tenant_id: str, page_params: PageQueryParams[Chunk]
    ) -> PageResponse[Chunk]:
        pass

    @abstractmethod
    async def get_chunk_by_id(
        self, tenant_id: str, chunk_id: str, embedding_model_name: str
    ) -> Union[Chunk, None]:
        pass

    @abstractmethod
    async def delete_chunk_by_id(
        self, tenant_id: str, chunk_id: str, embedding_model_name: str
    ) -> Union[Chunk, None]:
        pass

    @abstractmethod
    async def delete_knowledge_chunk(
        self, tenant_id: str, knowledge_ids: List[str]
    ) -> Union[List[Chunk], None]:
        pass

    # =================== retrieval ===================
    @abstractmethod
    async def search_space_chunk_list(
        self,
        tenant_id: str,
        params: RetrievalBySpaceRequest,
    ) -> List[RetrievalChunk]:
        pass

    @abstractmethod
    async def search_knowledge_chunk_list(
        self,
        tenant_id: str,
        params: RetrievalByKnowledgeRequest,
    ) -> List[RetrievalChunk]:
        pass

    @abstractmethod
    async def retrieve(
        self,
        tenant_id: str,
        params: RetrievalRequest,
    ) -> List[RetrievalChunk]:
        pass

    # =================== task ===================
    @abstractmethod
    async def save_task_list(self, task_list: List[Task]) -> List[Task]:
        pass

    @abstractmethod
    async def update_task_list(self, task_list: List[Task]) -> None:
        pass

    @abstractmethod
    async def get_task_list(
        self, tenant_id: str, page_params: PageQueryParams[Task]
    ) -> PageResponse[Task]:
        pass

    @abstractmethod
    async def get_task_by_id(self, tenant_id: str, task_id: str) -> Union[Task, None]:
        pass

    @abstractmethod
    async def delete_task_by_id(
        self, tenant_id: str, task_id: str
    ) -> Union[Task, None]:
        pass

    @abstractmethod
    async def task_statistics(
        self, space_id: str, status: TaskStatus
    ) -> Union[dict[TaskStatus, int], int]:
        pass

    @abstractmethod
    async def delete_knowledge_task(
        self, tenant_id: str, knowledge_ids: List[str]
    ) -> Union[List[Task], None]:
        pass

    # =================== tenant ===================
    @abstractmethod
    async def save_tenant(self, tenant: Tenant) -> Union[Tenant, None]:
        pass

    @abstractmethod
    async def get_tenant_by_sk(self, secret_key: str) -> Union[Tenant, None]:
        pass

    @abstractmethod
    async def update_tenant(self, tenant: Tenant) -> Union[Tenant, None]:
        pass

    @abstractmethod
    async def validate_tenant_name(self, tenant_name: str) -> bool:
        pass

    @abstractmethod
    async def get_tenant_by_id(self, tenant_id: str) -> Union[Tenant, None]:
        pass

    @abstractmethod
    async def get_tenant_list(
        self, page_params: PageQueryParams[Tenant]
    ) -> PageResponse[Tenant]:
        pass

    @abstractmethod
    async def delete_tenant_by_id(self, tenant_id: str) -> Union[Tenant, None]:
        pass

    # =================== api-key ===================
    @abstractmethod
    async def get_api_key_by_value(self, key_value: str) -> Union[APIKey, None]:
        pass

    @abstractmethod
    async def get_api_key_by_id(
        self, tenant_id: str, key_id: str
    ) -> Union[APIKey, None]:
        pass

    @abstractmethod
    async def get_tenant_api_keys(
        self, tenant_id: str, page_params: PageQueryParams[APIKey]
    ) -> PageResponse[APIKey]:
        pass

    @abstractmethod
    async def save_api_key(self, create_params: APIKey) -> APIKey:
        pass

    @abstractmethod
    async def update_api_key(self, update_params: APIKey) -> Union[APIKey, None]:
        pass

    @abstractmethod
    async def delete_api_key(self, key_id: str) -> bool:
        pass

    @abstractmethod
    async def get_all_expired_api_keys(self, tenant_id: str) -> List[APIKey]:
        pass

    # =================== rule ===================
    @abstractmethod
    async def get_tenant_rule(self, tenant_id: str) -> Optional[str]:
        pass

    @abstractmethod
    async def get_space_rule(self, tenant_id: str, space_id: str) -> Optional[str]:
        pass

    # =================== agent ===================
    @abstractmethod
    async def agent_invoke(self, params: Any) -> AsyncIterator[Any]:
        pass

    # =================== tag ===================
    @abstractmethod
    async def add_tag_list(
        self, tenant_id: str, tag_list: List[TagCreate]
    ) -> List[Tag]:
        pass

    @abstractmethod
    async def get_tag_list(
        self, tenant_id: str, page_params: PageQueryParams[Tag]
    ) -> PageResponse[Tag]:
        pass

    @abstractmethod
    async def get_tag_by_id(self, tenant_id: str, tag_id: str) -> Union[Tag, None]:
        pass

    @abstractmethod
    async def delete_tag_by_id(self, tenant_id: str, tag_id: str) -> Union[Tag, None]:
        pass

    @abstractmethod
    async def update_tag_name_description(
        self,
        tenant_id: str,
        tag_id: str,
        name: Optional[str],
        description: Optional[str],
    ) -> Tag:
        pass

    # =================== tagging ===================
    @abstractmethod
    async def add_tagging_list(
        self, tenant_id: str, tagging_list: List[TaggingCreate]
    ) -> List[Tagging]:
        pass

    @abstractmethod
    async def get_tagging_list(
        self, tenant_id: str, page_params: PageQueryParams[Tagging]
    ) -> PageResponse[Tagging]:
        pass

    @abstractmethod
    async def delete_tagging_by_id(
        self, tenant_id: str, tagging_id: str
    ) -> Union[Tagging, None]:
        pass

    # =================== artifacts ===================
    """
    ArtifactIndex 仓储层抽象接口
    """

    @abstractmethod
    async def add_artifact_list(
        self, artifact_list: List[ArtifactIndexCreate]
    ) -> List[ArtifactIndex]:
        """
        批量添加 artifact 记录
        返回实际插入成功的 ArtifactIndex 记录列表
        """
        pass

    @abstractmethod
    async def get_artifact_list(
        self, page_params: PageQueryParams[ArtifactIndex]
    ) -> PageResponse[ArtifactIndex]:
        """
        分页获取 artifact 列表
        """
        pass

    @abstractmethod
    async def get_artifact_by_id(self, artifact_id: str) -> Union[ArtifactIndex, None]:
        """
        根据 artifact_id 获取 ArtifactIndex 详情
        """
        pass

    @abstractmethod
    async def delete_artifact_by_id(
        self, artifact_id: str
    ) -> Union[ArtifactIndex, None]:
        """
        根据 artifact_id 删除 ArtifactIndex 记录
        返回被删除的记录（如果存在的话）
        """
        pass

    @abstractmethod
    async def update_artifact_space_id(
        self, artifact_id: str, new_space_id: str
    ) -> Union[ArtifactIndex, None]:
        """
        更新 artifact 的 space_id 绑定
        返回更新后的 ArtifactIndex 记录（如果存在的话）
        """
        pass

    # =================== dashboard ===================
    # TODO: add dashboard related methods

    # =================== webhook ===================
    @abstractmethod
    async def handle_webhook(
        self,
        tenant: Tenant,
        # webhook type, e.g. knowledge, chunk, etc.
        webhook_type: str,
        # webhook source, e.g. github, yuque, slack, etc.
        source: str,
        # knowledge base id
        knowledge_base_id: str,
        # webhook payload
        payload: Any,
    ) -> Optional[str]:
        pass



from typing import List

from whiskerrag_types.interface.parser_interface import BaseParser, ParseResult
from whiskerrag_types.model.knowledge import Knowledge
from whiskerrag_types.model.multi_modal import Image
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.PARSER, "base_image")
class BaseTextParser(BaseParser[Image]):

    async def parse(
        self,
        knowledge: Knowledge,
        content: Image,
    ) -> ParseResult:
        return [content]

    async def batch_parse(
        self,
        knowledge: Knowledge,
        content_list: List[Image],
    ) -> List[ParseResult]:
        return [await self.parse(knowledge, content) for content in content_list]



# tracing_context.py
from contextvars import ContextVar

# 定义全局 ContextVars
trace_id_ctx: ContextVar[str] = ContextVar("trace_id", default="unknown")
user_id_ctx: ContextVar[str] = ContextVar("user_id", default="unknown")
tenant_id_ctx: ContextVar[str] = ContextVar("tenant_id", default="system")


# --- traceId ---
def set_trace_id(trace_id: str) -> None:
    trace_id_ctx.set(trace_id)


def get_trace_id() -> str:
    return trace_id_ctx.get()


# --- userId ---
def set_user_id(user_id: str) -> None:
    user_id_ctx.set(user_id)


def get_user_id() -> str:
    return user_id_ctx.get()


# --- tenantId ---
def set_tenant_id(tenant_id: str) -> None:
    tenant_id_ctx.set(tenant_id)


def get_tenant_id() -> str:
    return tenant_id_ctx.get()



from typing import Any, AsyncIterator, Dict, List, Union

from openai import BaseModel

from whiskerrag_types.interface.llm_interface import BaseLLM, ContentType
from whiskerrag_utils.registry import RegisterTypeEnum, register


class ExampleResponse(BaseModel):
    """示例响应模型"""

    content: str
    model: str = "example-llm"
    usage: Dict[str, int] = {"tokens": 0}


@register(RegisterTypeEnum.LLM, "example", order=1)
class ExampleLLM(BaseLLM[ExampleResponse]):
    """示例LLM实现，用于演示注册系统"""

    def __init__(self, model_name: str = "example-llm"):
        self.model_name = model_name

    async def chat(
        self, content: Union[str, List[ContentType]], **kwargs: Any
    ) -> ExampleResponse:
        """示例聊天实现"""
        if isinstance(content, str):
            text_content = content
        else:
            # 处理混合内容
            text_parts = []
            for item in content:
                if isinstance(item, str):
                    text_parts.append(item)
                elif isinstance(item, dict) and item.get("type") == "image":
                    text_parts.append("[Image detected]")
            text_content = " ".join(text_parts)

        # 模拟响应
        response_text = f"Echo: {text_content}"

        return ExampleResponse(
            content=response_text,
            model=self.model_name,
            usage={"tokens": len(text_content.split())},
        )

    async def stream_chat(
        self, content: Union[str, List[ContentType]], **kwargs: Any
    ) -> AsyncIterator[ExampleResponse]:
        """示例流式聊天实现"""
        response = await self.chat(content, **kwargs)

        # 模拟流式输出，将响应分割为单词
        words = response.content.split()
        for i, word in enumerate(words):
            yield ExampleResponse(
                content=word + (" " if i < len(words) - 1 else ""),
                model=self.model_name,
                usage={"tokens": i + 1},
            )

    @classmethod
    async def health_check(cls) -> bool:
        """异步健康检查"""
        try:
            instance = cls()
            test_response = await instance.chat("test")
            return isinstance(test_response, ExampleResponse)
        except Exception as e:
            print(f"Health check failed: {e}")
            return False



import re
from typing import Any, Dict, List, Optional, Union

from whiskerrag_types.interface.loader_interface import BaseLoader
from whiskerrag_types.model.knowledge import (
    Knowledge,
    KnowledgeSourceEnum,
    KnowledgeTypeEnum,
)
from whiskerrag_types.model.knowledge_source import (
    OpenUrlSourceConfig,
    YuqueSourceConfig,
)
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_types.model.splitter import ImageSplitConfig
from whiskerrag_utils.helper.yuque import ExtendedYuqueLoader
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.KNOWLEDGE_LOADER, KnowledgeSourceEnum.YUQUE)
class WhiskerYuqueLoader(BaseLoader[Text]):
    _book_docs_cache: Optional[list]
    _doc_detail_cache: Dict[Union[str, int], Any]
    _yuque_loader: Optional[ExtendedYuqueLoader]

    def __init__(self, knowledge: Knowledge) -> None:
        super().__init__(knowledge)
        self._book_docs_cache = None
        self._doc_detail_cache = {}
        self._yuque_loader = None

    @property
    def yuque_loader(self) -> ExtendedYuqueLoader:
        if self._yuque_loader is None:
            if not isinstance(self.knowledge.source_config, YuqueSourceConfig):
                raise AttributeError("Invalid source config type for YuqueLoader")
            access_token = self.knowledge.source_config.auth_info
            api_url = self.knowledge.source_config.api_url
            self._yuque_loader = ExtendedYuqueLoader(
                access_token=access_token,
                api_url=api_url,
            )
        return self._yuque_loader

    def _create_doc_knowledge(self, parsed_document: Any) -> Knowledge:
        from copy import deepcopy

        source_config = deepcopy(self.knowledge.source_config)
        if hasattr(source_config, "document_id"):
            source_config.document_id = parsed_document.metadata.get(
                "slug"
            ) or parsed_document.metadata.get("id")
        return Knowledge(
            source_type=self.knowledge.source_type,
            knowledge_type=self.knowledge.knowledge_type,
            knowledge_name=f"{self.knowledge.knowledge_name}/{parsed_document.metadata.get('title', '')}",
            embedding_model_name=self.knowledge.embedding_model_name,
            source_config=source_config,
            tenant_id=self.knowledge.tenant_id,
            file_size=len(parsed_document.page_content.encode("utf-8")),
            file_sha=parsed_document.metadata.get("content_updated_at"),
            space_id=self.knowledge.space_id,
            split_config=self.knowledge.split_config,
            parent_id=self.knowledge.knowledge_id,
            enabled=True,
            metadata=parsed_document.metadata,
        )

    def _create_image_knowledges(self, parsed_document: Any) -> List[Knowledge]:
        image_knowledges = []
        image_pattern = r"!\[(.*?)\]\((.*?)\)"
        all_image_matches = re.findall(image_pattern, parsed_document.page_content)

        for img_idx, (alt_text, img_url) in enumerate(all_image_matches):
            if img_url.strip():
                img_metadata = parsed_document.metadata.copy()
                img_metadata["_img_idx"] = img_idx
                img_metadata["_img_url"] = img_url.strip()
                img_metadata["_alt_text"] = alt_text.strip() if alt_text.strip() else ""

                image_knowledge = Knowledge(
                    source_type=KnowledgeSourceEnum.CLOUD_STORAGE_IMAGE,
                    knowledge_type=KnowledgeTypeEnum.IMAGE,
                    knowledge_name=f"{self.knowledge.knowledge_name}/{parsed_document.metadata.get('title', '')}/image_{img_idx}",
                    embedding_model_name=self.knowledge.embedding_model_name,
                    source_config=OpenUrlSourceConfig(url=img_url.strip()),
                    tenant_id=self.knowledge.tenant_id,
                    space_id=self.knowledge.space_id,
                    split_config=ImageSplitConfig(),
                    parent_id=self.knowledge.knowledge_id,
                    enabled=True,
                    metadata=img_metadata,
                )
                image_knowledges.append(image_knowledge)
        return image_knowledges

    def get_doc_detail(
        self, group_login: str, book_slug: str, document_id: Optional[Union[str, int]]
    ) -> Any:
        if document_id in self._doc_detail_cache:
            return self._doc_detail_cache[document_id]
        if document_id is None:
            raise ValueError("document_id is required")
        parsed_document = self.yuque_loader.load_document_by_path(
            group_login, book_slug, document_id
        )
        self._doc_detail_cache[document_id] = parsed_document
        return parsed_document

    def get_book_docs(self, group_login: str, book_slug: str) -> list:
        if self._book_docs_cache is not None:
            return self._book_docs_cache
        docs = self.yuque_loader.get_book_documents_by_path(group_login, book_slug)
        self._book_docs_cache = docs
        return docs

    async def load(self) -> List[Text]:
        if not isinstance(self.knowledge.source_config, YuqueSourceConfig):
            raise AttributeError("Invalid source config type for YuqueLoader")
        group_login = self.knowledge.source_config.group_login
        book_slug = self.knowledge.source_config.book_slug
        document_id = self.knowledge.source_config.document_id
        text_list: List[Text] = []
        try:
            if not group_login:
                raise ValueError("group_login is needed for WhiskerYuqueLoader")
            if not book_slug:
                raise ValueError("book_slug is needed for WhiskerYuqueLoader")
            if not document_id:
                raise ValueError("document_id is needed for WhiskerYuqueLoader")
            parsed_document = self.get_doc_detail(group_login, book_slug, document_id)
            text_list.append(
                Text(
                    content=parsed_document.page_content,
                    metadata=parsed_document.metadata,
                )
            )
            return text_list
        except Exception as e:
            raise Exception(f"Failed to load content from Yuque: {e}")

    async def decompose(self) -> List[Knowledge]:
        if not isinstance(self.knowledge.source_config, YuqueSourceConfig):
            raise AttributeError("Invalid source config type for YuqueLoader")

        group_login = self.knowledge.source_config.group_login
        book_slug = self.knowledge.source_config.book_slug
        document_id = self.knowledge.source_config.document_id

        if not group_login:
            raise ValueError("group_login is needed for WhiskerYuqueLoader")
        if not book_slug:
            raise ValueError("book_slug is needed for WhiskerYuqueLoader")

        knowledge_list: List[Knowledge] = []

        # Decompose a specific document - only extract images
        if document_id:
            doc_knowledge_list = await self._decompose_document(
                group_login, book_slug, document_id, include_document=False
            )
            knowledge_list.extend(doc_knowledge_list)
        # Decompose a whole book - create sub-documents as knowledge, no images
        else:
            book_knowledges = await self._decompose_book(group_login, book_slug)
            knowledge_list.extend(book_knowledges)

        return knowledge_list

    async def _decompose_book(
        self, group_login: str, book_slug: str
    ) -> List[Knowledge]:
        knowledges: List[Knowledge] = []
        try:
            docs = self.get_book_docs(group_login, book_slug)
            for doc in docs:
                doc_knowledges = await self._decompose_document(
                    group_login,
                    book_slug,
                    doc["slug"],
                    include_document=True,
                    include_images=False,
                )
                knowledges.extend(doc_knowledges)
        except Exception as e:
            raise ValueError(f"Failed to get book documents: {e}")
        return knowledges

    async def _decompose_document(
        self,
        group_login: str,
        book_slug: str,
        document_id: Optional[Union[str, int]],
        include_document: bool = True,
        include_images: bool = True,
    ) -> List[Knowledge]:
        knowledges: List[Knowledge] = []
        try:
            parsed_document = self.get_doc_detail(group_login, book_slug, document_id)

            # Only create document knowledge if include_document is True
            if include_document:
                doc_knowledge = self._create_doc_knowledge(parsed_document)
                knowledges.append(doc_knowledge)

            # Only create image knowledges if include_images is True
            if include_images:
                image_knowledges = self._create_image_knowledges(parsed_document)
                knowledges.extend(image_knowledges)
        except Exception as e:
            raise ValueError(f"Failed to get document: {e}")
        return knowledges

    async def on_load_finished(self) -> None:
        pass



import json
import os
import re
import xml.etree.ElementTree as ET
from configparser import ConfigParser
from pathlib import Path
from typing import Dict, List, Optional, Tuple

try:
    import tomllib  # type: ignore
except ImportError:
    import tomli as tomllib


# ===== Python =====
def parse_python_config(dir_path: Path) -> Optional[Tuple[str, str]]:
    pyproject = dir_path / "pyproject.toml"
    if pyproject.is_file():
        try:
            with pyproject.open("rb") as f:
                data = tomllib.load(f)
            if "project" in data:
                name = data["project"].get("name")
                version = data["project"].get("version")
                if name and version:
                    return name, version
            if "tool" in data and "poetry" in data["tool"]:
                name = data["tool"]["poetry"].get("name")
                version = data["tool"]["poetry"].get("version")
                if name and version:
                    return name, version
        except Exception:
            pass

    setup_cfg = dir_path / "setup.cfg"
    if setup_cfg.is_file():
        try:
            config = ConfigParser()
            config.read(setup_cfg)
            if config.has_section("metadata"):
                name = config.get("metadata", "name", fallback=None)
                version = config.get("metadata", "version", fallback=None)
                if name and version:
                    return name, version
        except Exception:
            pass

    setup_py = dir_path / "setup.py"
    if setup_py.is_file():
        try:
            content = setup_py.read_text(encoding="utf-8", errors="ignore")
            name_match = re.search(r"name\s*=\s*['\"]([A-Za-z0-9_.\-]+)['\"]", content)
            ver_match = re.search(
                r"version\s*=\s*['\"]([A-Za-z0-9_.\-]+)['\"]", content
            )
            if name_match and ver_match:
                return name_match.group(1), ver_match.group(1)
        except Exception:
            pass
    return None


# ===== Node.js =====
def parse_node_config(dir_path: Path) -> Optional[Tuple[str, str]]:
    package_json = dir_path / "package.json"
    if package_json.is_file():
        try:
            with package_json.open(encoding="utf-8") as f:
                data = json.load(f)
            name = data.get("name")
            version = data.get("version")
            if name and version:
                return name, version
        except Exception:
            pass
    return None


# ===== Maven =====
def parse_maven_config(dir_path: Path) -> Optional[Tuple[str, str]]:
    pom_xml = dir_path / "pom.xml"
    if pom_xml.is_file():
        try:
            tree = ET.parse(pom_xml)
            root = tree.getroot()
            ns_match = re.match(r"\{.*\}", root.tag)
            nsmap = {"mvn": ns_match.group(0)[1:-1]} if ns_match else {}

            def find_tag(tag: str) -> Optional[str]:
                if nsmap:
                    el = root.find(f"mvn:{tag}", namespaces=nsmap)
                else:
                    el = root.find(tag)
                return el.text.strip() if el is not None and el.text else None

            artifact_id = find_tag("artifactId")
            version = find_tag("version")
            if artifact_id and version:
                return artifact_id, version
        except Exception:
            pass
    return None


# ===== Go =====
def parse_go_config(dir_path: Path) -> Optional[Tuple[str, str]]:
    go_mod = dir_path / "go.mod"
    if go_mod.is_file():
        try:
            with go_mod.open(encoding="utf-8") as f:
                for line in f:
                    line = line.strip()
                    if line.startswith("module "):
                        module_path = line.split()[1]
                        name = module_path.split("/")[-1]  # 模块名用最后一段
                        version = ""  # go.mod 通常不含自身版本
                        return name, version
        except Exception:
            pass
    return None


# ===== PHP =====
def parse_php_config(dir_path: Path) -> Optional[Tuple[str, str]]:
    composer_json = dir_path / "composer.json"
    if composer_json.is_file():
        try:
            with composer_json.open(encoding="utf-8") as f:
                data = json.load(f)
            name = data.get("name")
            version = data.get("version", "")
            if name:
                return name, version
        except Exception:
            pass
    return None


# ===== 主方法 =====
def get_all_build_artifacts(repo_path: str) -> List[Dict]:
    repo = Path(repo_path).resolve()
    results = []

    for root, dirs, files in os.walk(repo):
        dir_path = Path(root)
        rel_path = os.path.relpath(dir_path, repo)  # 相对于 repo_path 的路径
        if rel_path == ".":
            rel_path = "."  # 根目录

        for parser in (
            parse_python_config,
            parse_node_config,
            parse_maven_config,
            parse_go_config,
            parse_php_config,
        ):
            info = parser(dir_path)
            if info:
                results.append(
                    {"path": rel_path, "name": info[0], "version": info[1]}  # 相对路径
                )
    return results


# ==== 测试 ====
if __name__ == "__main__":
    repo_dir = "/path/to/your/repo"
    artifacts = get_all_build_artifacts(repo_dir)
    for art in artifacts:
        print(f"[{art['path']}] 产物名: {art['name']}, 版本号: {art['version']}")



import hashlib
import json
import logging
import os
import shutil
import tempfile
import time
import uuid
import zipfile
from contextlib import contextmanager
from threading import Lock
from typing import Any, Dict, Iterator, Optional, Union

from whiskerrag_types.model.knowledge_source import GithubRepoSourceConfig

from .utils import log_system_info

logger = logging.getLogger(__name__)


# ===== Mock 类型（ZIP 模式下 GitPython 无法工作） =====
class MockCommit:
    def __init__(
        self,
        sha: str,
        author_name: str = "Unknown",
        author_email: str = "unknown@example.com",
        message: str = "Mock commit",
    ) -> None:
        self.hexsha = sha
        self.author = MockAuthor(author_name, author_email)
        self.message = message
        self.tree: Optional[MockTree] = None


class MockAuthor:
    def __init__(self, name: str, email: str):
        self.name = name
        self.email = email


class MockTreeItem:
    def __init__(self, full_path: str, relative_path: str, item_type: str):
        self.full_path = full_path
        self.path = relative_path
        self.type = item_type  # "blob" / "tree"


class MockTree:
    def __init__(self, repo_path: str = ""):
        self.repo_path = repo_path

    def __truediv__(self, path: str) -> "MockBlob":
        return MockBlob(os.path.join(self.repo_path, path), path)

    def traverse(self) -> Iterator[MockTreeItem]:
        if not os.path.exists(self.repo_path):
            return
        for root, dirs, files in os.walk(self.repo_path):
            if ".git" in dirs:
                dirs.remove(".git")
            for d in dirs:
                dp = os.path.join(root, d)
                yield MockTreeItem(dp, os.path.relpath(dp, self.repo_path), "tree")
            for f in files:
                fp = os.path.join(root, f)
                yield MockTreeItem(fp, os.path.relpath(fp, self.repo_path), "blob")


class MockBlob:
    def __init__(self, file_path: str, relative_path: str):
        self.file_path = file_path
        self.relative_path = relative_path
        self.path = relative_path
        self.type = "blob"

    @property
    def hexsha(self) -> str:
        try:
            with open(self.file_path, "rb") as f:
                return hashlib.sha1(f.read()).hexdigest()
        except Exception:
            return "0" * 40

    @property
    def size(self) -> int:
        try:
            return os.path.getsize(self.file_path)
        except Exception:
            return 0

    @property
    def mode(self) -> int:
        try:
            return os.stat(self.file_path).st_mode
        except Exception:
            return 0o100644


class MockHead:
    def __init__(self, commit: MockCommit):
        self.commit = commit


class MockBranch:
    def __init__(self, name: str):
        self.name = name


class MockRepo:
    def __init__(
        self, repo_path: str, config: GithubRepoSourceConfig, repo_info: Dict[str, Any]
    ):
        self.working_dir = repo_path
        self.repo_path = repo_path
        self.config = config
        self.repo_info = repo_info
        sha = repo_info.get("default_branch_sha", "unknown")
        author_name = repo_info.get("owner", {}).get("login", "Unknown")
        self._commit = MockCommit(sha, author_name)
        self._commit.tree = MockTree(repo_path)
        self.head = MockHead(self._commit)
        branch_name = config.branch or repo_info.get("default_branch", "main")
        self.active_branch = MockBranch(branch_name)

    def iter_commits(self, **kwargs: Any) -> Iterator[MockCommit]:
        yield self._commit


# ===== 全局锁表（防止相同 repo_key 并发冲突） =====
_repo_locks: dict[str, Lock] = {}


@contextmanager
def repo_lock(key: str) -> Any:
    lock = _repo_locks.setdefault(key, Lock())
    lock.acquire()
    try:
        yield
    finally:
        lock.release()


# ===== Git 仓库管理器 =====
def _check_git_installation() -> bool:
    try:
        import subprocess

        subprocess.run(["git", "--version"], check=True, capture_output=True, text=True)
        return True
    except (subprocess.SubprocessError, FileNotFoundError):
        return False


def _lazy_import_git() -> Any:
    try:
        from git import Repo
        from git.exc import GitCommandNotFound, InvalidGitRepositoryError

        return Repo, GitCommandNotFound, InvalidGitRepositoryError
    except ImportError as e:
        raise ImportError(
            "GitPython is required, install with `pip install GitPython`"
        ) from e


def _lazy_import_requests() -> Any:
    try:
        import requests

        return requests
    except ImportError as e:
        raise ImportError(
            "requests is required, install with `pip install requests`"
        ) from e


class GitRepoManager:
    CACHE_TTL_HOURS = 24  # 缓存有效时间

    def __init__(self) -> None:
        self.repos_dir = os.environ.get("WHISKER_REPO_SAVE_PATH") or os.path.join(
            tempfile.gettempdir(), "repo_download"
        )
        self._repos_cache: Dict[str, str] = {}
        self._repo_info_cache: Dict[str, Dict[str, Any]] = {}
        os.makedirs(self.repos_dir, exist_ok=True)

    # ==== 缓存时间标记 ====
    def _is_cache_expired(self, repo_path: str) -> bool:
        meta_path = os.path.join(repo_path, ".whisker_meta")
        if not os.path.exists(meta_path):
            return True
        try:
            with open(meta_path, "r") as f:
                last_time = float(json.load(f).get("last_update", 0))
            return (time.time() - last_time) > self.CACHE_TTL_HOURS * 3600
        except Exception:
            return True

    def _mark_cache_updated(self, repo_path: str) -> None:
        try:
            with open(os.path.join(repo_path, ".whisker_meta"), "w") as f:
                json.dump({"last_update": time.time()}, f)
        except Exception as e:
            logger.warning(f"Failed to mark cache update: {e}")

    def _safe_rmtree(self, path: str) -> None:
        try:
            shutil.rmtree(path)
        except OSError as e:
            logger.warning(f"Safe cleanup failed for {path}: {e}, trying rename")
            try:
                renamed = f"{path}_old_{uuid.uuid4().hex[:6]}"
                os.rename(path, renamed)
            except Exception as e2:
                logger.error(f"Rename also failed: {e2}")

    def _generate_repo_key(self, config: GithubRepoSourceConfig) -> str:
        parts = [config.repo_name]
        if config.branch:
            parts.append(config.branch)
        if config.commit_id:
            parts.append(config.commit_id)
        return "_".join(parts).replace("/", "_")

    def _is_github_repo(self, config: GithubRepoSourceConfig) -> bool:
        return "github.com" in config.url.lower()

    # ==== 核心 API ====
    def get_repo_path(self, config: GithubRepoSourceConfig) -> str:
        repo_key = self._generate_repo_key(config)
        with repo_lock(repo_key):
            repo_path = os.path.join(self.repos_dir, repo_key)
            if os.path.exists(repo_path) and not self._is_cache_expired(repo_path):
                logger.info(f"Using cached repository: {repo_path}")
                self._repos_cache[repo_key] = repo_path
                return repo_path
            if os.path.exists(repo_path):
                self._safe_rmtree(repo_path)
            repo_path = self._download_repo(config, repo_path)
            self._repos_cache[repo_key] = repo_path
            return repo_path

    def get_repo(self, config: GithubRepoSourceConfig) -> Union[Any, MockRepo]:
        """返回 GitPython Repo 或 MockRepo"""
        repo_path = self.get_repo_path(config)
        git_dir = os.path.join(repo_path, ".git")
        if os.path.exists(git_dir):
            try:
                Repo, _, _ = _lazy_import_git()
                return Repo(repo_path)
            except Exception as e:
                logger.warning(f"Failed to create GitPython Repo: {e}")
        if self._is_github_repo(config):
            repo_key = self._generate_repo_key(config)
            return MockRepo(repo_path, config, self._repo_info_cache.get(repo_key, {}))
        raise ValueError("Repository is not a git repo and not on GitHub")

    def cleanup_repo(self, config: GithubRepoSourceConfig) -> None:
        """
        清理指定仓库缓存目录和缓存信息
        """
        repo_key = self._generate_repo_key(config)
        if repo_key in self._repos_cache:
            repo_path = self._repos_cache[repo_key]
            try:
                if os.path.exists(repo_path):
                    self._safe_rmtree(repo_path)
                    logger.info(f"Cleaned up repository: {repo_path}")
            except Exception as e:
                logger.error(f"Error cleaning up repository {repo_path}: {e}")
            finally:
                del self._repos_cache[repo_key]
        if repo_key in self._repo_info_cache:
            del self._repo_info_cache[repo_key]

    # ==== 下载逻辑 ====
    def _download_repo(
        self, config: GithubRepoSourceConfig, repo_saved_path: str
    ) -> str:
        if os.path.exists(repo_saved_path):
            self._safe_rmtree(repo_saved_path)
        if _check_git_installation():
            try:
                clone_url = self._build_clone_url(config)
                self._clone_repo(
                    clone_url, repo_saved_path, config.branch, config.commit_id
                )
                self._mark_cache_updated(repo_saved_path)
                return repo_saved_path
            except Exception as e:
                logger.warning(f"Git clone failed: {e}")
                log_system_info()
                if os.path.exists(repo_saved_path):
                    self._safe_rmtree(repo_saved_path)
        if self._is_github_repo(config):
            self._download_github_zip(config, repo_saved_path)
            self._mark_cache_updated(repo_saved_path)
            return repo_saved_path
        raise ValueError(f"Failed to download repository: {config.repo_name}")

    def _build_clone_url(self, config: GithubRepoSourceConfig) -> str:
        base_url = config.url.rstrip("/")
        repo_name = config.repo_name
        if base_url.startswith("https://"):
            url_no_scheme = base_url[8:]
        elif base_url.startswith("http://"):
            url_no_scheme = base_url[7:]
        else:
            url_no_scheme = base_url
        if config.auth_info:
            return f"https://{config.auth_info}@{url_no_scheme}/{repo_name}.git"
        return f"https://{url_no_scheme}/{repo_name}.git"

    def _clone_repo(
        self,
        clone_url: str,
        repo_path: str,
        branch: Optional[str] = None,
        commit_id: Optional[str] = None,
        initial_depth: int = 1,
        max_fetch_tries: int = 5,
        depth_step: int = 20,
        max_retry_attempts: int = 3,
        retry_delay: float = 2.0,
    ) -> None:
        Repo, GitCommandNotFound, InvalidGitRepositoryError = _lazy_import_git()
        last_exception = None
        import subprocess

        for attempt in range(max_retry_attempts):
            try:
                if attempt > 0 and os.path.exists(repo_path):
                    self._safe_rmtree(repo_path)
                logger.info(
                    f"Cloning repository (attempt {attempt + 1}/{max_retry_attempts}):{clone_url}"
                )
                if branch:
                    repo = Repo.clone_from(
                        clone_url,
                        repo_path,
                        multi_options=["--filter=blob:limit=5m"],
                        branch=branch,
                        depth=initial_depth,
                    )
                else:
                    repo = Repo.clone_from(
                        clone_url,
                        repo_path,
                        multi_options=["--filter=blob:limit=5m"],
                        depth=initial_depth,
                    )

                # 放开 git 安全校验
                try:
                    subprocess.run(
                        [
                            "git",
                            "config",
                            "--global",
                            "--add",
                            "safe.directory",
                            repo_path,
                        ],
                        check=False,
                    )
                except Exception as se:
                    logger.warning(f"Could not set safe.directory for git: {se}")

                if commit_id:
                    for i in range(max_fetch_tries):
                        try:
                            repo.git.checkout(commit_id)
                            break
                        except Exception as e:
                            if any(
                                k in str(e).lower()
                                for k in ["not found", "unknown revision", "pathspec"]
                            ):
                                repo.git.fetch("origin", f"--deepen={depth_step}")
                                if branch:
                                    repo.git.checkout(branch)
                                continue
                            else:
                                raise
                    else:
                        raise ValueError(f"Failed to fetch commit {commit_id}")
                self._mark_cache_updated(repo_path)
                return
            except GitCommandNotFound:
                raise ValueError("Git command not found.")
            except InvalidGitRepositoryError as e:
                raise ValueError(f"Invalid git repository: {e}")
            except Exception as e:
                last_exception = e
                if attempt < max_retry_attempts - 1:
                    time.sleep(retry_delay)
                    retry_delay *= 2

        if last_exception:
            raise ValueError(
                f"Failed to clone after {max_retry_attempts} attempts: {last_exception}"
            )

    def _download_github_zip(
        self, config: GithubRepoSourceConfig, repo_path: str
    ) -> None:
        requests = _lazy_import_requests()
        repo_info = self._get_github_repo_info(config)
        ref = (
            config.commit_id or config.branch or repo_info.get("default_branch", "main")
        )
        zip_url = f"https://api.github.com/repos/{config.repo_name}/zipball/{ref}"
        headers = {}
        if config.auth_info:
            headers["Authorization"] = f"token {config.auth_info}"
        resp = requests.get(zip_url, headers=headers, stream=True)
        resp.raise_for_status()
        zip_path = os.path.join(
            self.repos_dir, f"{config.repo_name.replace('/', '_')}_{ref}.zip"
        )
        with open(zip_path, "wb") as f:
            for chunk in resp.iter_content(8192):
                f.write(chunk)
        try:
            if os.path.exists(repo_path):
                self._safe_rmtree(repo_path)
            with zipfile.ZipFile(zip_path, "r") as zf:
                zf.extractall(self.repos_dir)
                part = config.repo_name.split("/")[-1]
                candidates = [
                    d
                    for d in os.listdir(self.repos_dir)
                    if os.path.isdir(os.path.join(self.repos_dir, d))
                ]
                for d in sorted(candidates):
                    if part in d:
                        shutil.move(os.path.join(self.repos_dir, d), repo_path)
                        return
                raise ValueError(f"Could not find extracted repo for {part}")
        finally:
            if os.path.exists(zip_path):
                os.unlink(zip_path)

    def _get_github_repo_info(self, config: GithubRepoSourceConfig) -> Dict[str, Any]:
        requests = _lazy_import_requests()
        repo_key = self._generate_repo_key(config)
        if repo_key in self._repo_info_cache:
            return self._repo_info_cache[repo_key]
        api_url = f"https://api.github.com/repos/{config.repo_name}"
        headers = {}
        if config.auth_info:
            headers["Authorization"] = f"token {config.auth_info}"
        resp = requests.get(api_url, headers=headers)
        resp.raise_for_status()
        info: Dict[str, Any] = resp.json()
        ref = config.commit_id or config.branch or info.get("default_branch", "main")
        branch_url = f"https://api.github.com/repos/{config.repo_name}/branches/{ref}"
        br_resp = requests.get(branch_url, headers=headers)
        if br_resp.status_code == 200:
            info["default_branch_sha"] = br_resp.json()["commit"]["sha"]
        else:
            info["default_branch_sha"] = "unknown"
        self._repo_info_cache[repo_key] = info
        return info


# ===== 提供全局单例 =====
_repo_manager = GitRepoManager()


def get_repo_manager() -> GitRepoManager:
    return _repo_manager



import asyncio
import logging
import uuid
from typing import Any, Dict, List, Optional, Tuple, TypedDict, Union

from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.knowledge import Knowledge, KnowledgeTypeEnum
from whiskerrag_types.model.multi_modal import Image, Text

from .registry import (
    RegisterTypeEnum,
    get_all_registered_with_metadata,
    get_register,
    get_register_metadata,
    get_register_order,
    init_register,
    register,
)

logger = logging.getLogger("whisker")


class DiffResult(TypedDict):
    to_add: List[Knowledge]
    to_delete: List[Knowledge]
    unchanged: List[Knowledge]


def _process_metadata_and_tags(
    knowledge: Knowledge, parse_item: Union[Text, Image]
) -> Tuple[Dict[str, Any], List[str]]:
    """
    Process metadata and tags for a parse item.
    Args:
        knowledge: The knowledge object.
        parse_item: The parse item (Text or Image).
    Returns:
        A tuple of (combined_metadata, tags).
    """
    combined_metadata = {**knowledge.metadata, **parse_item.metadata}
    combined_metadata["_knowledge_type"] = knowledge.knowledge_type
    combined_metadata["_reference_url"] = combined_metadata.get("_reference_url", "")

    # Extract tags from metadata
    tags = combined_metadata.get("_tags")
    if isinstance(tags, str):
        tags = [tag.strip() for tag in tags.split(",") if tag.strip()]
    elif not isinstance(tags, list):
        tags = []

    return combined_metadata, tags


def _create_chunk(
    knowledge: Knowledge,
    parse_item: Union[Text, Image],
    embedding: List[float],
    combined_metadata: Dict[str, Any],
    tags: List[str],
) -> Chunk:
    """
    Create a Chunk object from the given parameters.
    Args:
        knowledge: The knowledge object.
        parse_item: The parse item (Text or Image).
        embedding: The embedding vector.
        combined_metadata: The combined metadata.
        tags: The tags.
    Returns:
        A Chunk object.
    """
    return Chunk(
        chunk_id=str(uuid.uuid4()),
        space_id=knowledge.space_id,
        tenant_id=knowledge.tenant_id,
        knowledge_id=knowledge.knowledge_id,
        context=parse_item.content if isinstance(parse_item, Text) else "",
        embedding=embedding,
        enabled=knowledge.enabled,
        embedding_model_name=knowledge.embedding_model_name,
        metadata=combined_metadata,
        tags=tags,
        f1=combined_metadata.get("_f1"),
        f2=combined_metadata.get("_f2"),
        f3=combined_metadata.get("_f3"),
        f4=combined_metadata.get("_f4"),
        f5=combined_metadata.get("_f5"),
    )


def _get_unique_origin_list(
    origin_list: List[Knowledge],
) -> Tuple[List[Knowledge], List[Knowledge]]:
    to_delete = []
    seen_file_shas = set()
    unique_origin_list = []
    for item in origin_list:
        if item.file_sha not in seen_file_shas:
            seen_file_shas.add(item.file_sha)
            unique_origin_list.append(item)
        else:
            to_delete.append(item)
    return to_delete, unique_origin_list


async def decompose_knowledge(
    knowledge: Knowledge,
    semaphore: Optional[asyncio.Semaphore] = None,
    max_concurrency: int = 4,
) -> List[Knowledge]:
    if semaphore is None:
        semaphore = asyncio.Semaphore(max_concurrency)
    LoaderCls = get_register(RegisterTypeEnum.KNOWLEDGE_LOADER, knowledge.source_type)
    async with semaphore:
        knowledge_list = await LoaderCls(knowledge).decompose()
    if not knowledge_list:
        return []
    tasks = [decompose_knowledge(k, semaphore) for k in knowledge_list]
    results = await asyncio.gather(*tasks)
    flat = [item for sublist in results for item in sublist]
    return flat if flat else knowledge_list


async def get_chunks_by_knowledge(knowledge: Knowledge) -> List[Chunk]:
    """
    Convert knowledge into vectorized chunks with controlled concurrency
    """
    source_type = knowledge.source_type
    knowledge_type = knowledge.knowledge_type
    parse_type = getattr(
        knowledge.split_config,
        "type",
        "base_image" if knowledge_type is KnowledgeTypeEnum.IMAGE else "base_text",
    )
    LoaderCls = get_register(RegisterTypeEnum.KNOWLEDGE_LOADER, source_type)
    ParserCls = get_register(RegisterTypeEnum.PARSER, parse_type)
    EmbeddingCls = get_register(
        RegisterTypeEnum.EMBEDDING, knowledge.embedding_model_name
    )
    # If no parser, return empty list
    if ParserCls is None:
        logger.warning(f"No parser found for type: {parse_type}")
        return []
    # If no embedding model, return empty list
    if EmbeddingCls is None:
        logger.warning(
            f"[warn]: No embedding model found for name: {knowledge.embedding_model_name}"
        )
        return []
    loaded_contents = []
    if LoaderCls is None:
        # If no loader, directly parse the knowledge object itself
        logger.warning(
            f"No loader found for source type: {knowledge.source_type}, attempting to parse knowledge directly."
        )
        parse_results = await ParserCls().parse(knowledge, None)
    else:
        # Use loader to load contents
        loaded_contents = await LoaderCls(knowledge).load()
        if not loaded_contents:
            logger.warning(
                f"Loader returned no content for source type: {knowledge.source_type}."
            )
            return []
        # Parse loaded contents
        parse_results = []
        for content in loaded_contents:
            split_result = await ParserCls().parse(knowledge, content)
            parse_results.extend(split_result)

    # Classify parse_results by type
    text_items = []
    image_items = []
    for parse_item in parse_results:
        if isinstance(parse_item, Text):
            text_items.append(parse_item)
        elif isinstance(parse_item, Image):
            image_items.append(parse_item)
        else:
            logger.warning(f"[warn]: illegal parse item: {parse_item}")

    chunks = []

    # Batch process Text items using embed_documents
    if text_items:
        try:
            logger.info(f"Processing {len(text_items)} text items in batch")
            documents = [text_item.content for text_item in text_items]
            embeddings = await EmbeddingCls().embed_documents(documents, timeout=30)
            for text_item, embedding in zip(text_items, embeddings):
                combined_metadata, tags = _process_metadata_and_tags(
                    knowledge, text_item
                )
                chunk = _create_chunk(
                    knowledge, text_item, embedding, combined_metadata, tags
                )
                chunks.append(chunk)
        except Exception as e:
            logger.error(f"Error processing text items in batch: {e}")

    # Process Image items individually
    if image_items:
        for image_item in image_items:
            try:
                logger.info(f"Processing image item: {image_item}")
                embedding = await EmbeddingCls().embed_image(image_item, timeout=60 * 5)
                if not embedding:
                    logger.warning(
                        f"[warn]: embed image failed, image item: {image_item}"
                    )
                    continue
                combined_metadata, tags = _process_metadata_and_tags(
                    knowledge, image_item
                )
                chunk = _create_chunk(
                    knowledge, image_item, embedding, combined_metadata, tags
                )
                chunks.append(chunk)
            except Exception as e:
                logger.error(f"Error processing image item: {e}")
                continue

    return chunks


def get_diff_knowledge_by_sha(
    origin_list: Optional[List[Knowledge]] = None,
    new_list: Optional[List[Knowledge]] = None,
) -> DiffResult:
    try:
        origin_list = origin_list or []
        new_list = new_list or []

        to_delete = []
        to_delete_origin, unique_origin_list = _get_unique_origin_list(origin_list)
        to_delete.extend(to_delete_origin)
        _, unique_new_list = _get_unique_origin_list(new_list)

        origin_map = {item.file_sha: item for item in unique_origin_list}

        to_add = []
        unchanged = []
        for new_item in unique_new_list:
            if new_item.file_sha not in origin_map:
                to_add.append(new_item)
            else:
                unchanged.append(new_item)
                del origin_map[new_item.file_sha]

        to_delete.extend(list(origin_map.values()))

        return {"to_add": to_add, "to_delete": to_delete, "unchanged": unchanged}
    except Exception as error:
        logger.error(f"Error in _get_diff_knowledge_by_sha: {error}")
        return {"to_add": [], "to_delete": [], "unchanged": []}


__all__ = [
    "get_register",
    "get_register_metadata",
    "get_register_order",
    "get_all_registered_with_metadata",
    "register",
    "RegisterTypeEnum",
    "init_register",
    "decompose_knowledge",
    "get_chunks_by_knowledge",
    "DiffResult",
    "get_diff_knowledge_by_sha",
]



import json
from typing import Any, Dict, List

from langchain_text_splitters import RecursiveJsonSplitter

from whiskerrag_types.interface.parser_interface import BaseParser, ParseResult
from whiskerrag_types.model import JSONSplitConfig, Knowledge
from whiskerrag_types.model.knowledge import KnowledgeTypeEnum
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_utils.registry import RegisterTypeEnum, register


def parse_json_content(content: str) -> Dict[str, Any]:
    """
    Parse JSON content with better compatibility.

    Args:
        content: JSON string content

    Returns:
        Parsed JSON object as dictionary (arrays are converted to {"data": [...]})

    Raises:
        ValueError: If content is not valid JSON or is not a dict/list
    """
    try:
        json_content = json.loads(content)

        # Support both dict and list formats
        if isinstance(json_content, dict):
            return json_content
        elif isinstance(json_content, list):
            # Convert list to dict format for processing
            return {"data": json_content}
        else:
            raise ValueError("JSON content must be a dictionary or array.")

    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON content provided for splitting: {str(e)}")
    except Exception as e:
        raise ValueError(f"Error processing JSON content: {str(e)}")


@register(RegisterTypeEnum.PARSER, KnowledgeTypeEnum.JSON)
class JSONParser(BaseParser[Text]):
    async def parse(
        self,
        knowledge: Knowledge,
        content: Text,
    ) -> ParseResult:
        """Splits JSON content into smaller chunks based on the provided configuration."""
        split_config = knowledge.split_config
        if not isinstance(split_config, JSONSplitConfig):
            raise TypeError("knowledge.split_config must be of type JSONSplitConfig")

        # Use the new helper function for better JSON compatibility
        json_content = parse_json_content(content.content)

        splitter = RecursiveJsonSplitter(
            max_chunk_size=split_config.max_chunk_size,
            min_chunk_size=split_config.min_chunk_size,
        )
        split_texts = splitter.split_text(
            json_content, convert_lists=True, ensure_ascii=False
        )
        result: ParseResult = []
        for idx, text in enumerate(split_texts):
            metadata = content.metadata.copy()
            metadata["_idx"] = idx
            result.append(Text(content=text, metadata=metadata))
        return result

    async def batch_parse(
        self,
        knowledge: Knowledge,
        content_list: List[Text],
    ) -> List[ParseResult]:
        return [await self.parse(knowledge, content) for content in content_list]



import re
from datetime import datetime, timezone
from enum import Enum
from typing import Any
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator, model_validator

from whiskerrag_types.model.timeStampedModel import TimeStampedModel


class TagObjectType(str, Enum):
    """
    可打标签的对象类型。
    """

    SPACE = "space"
    KNOWLEDGE = "knowledge"


class TaggingCreate(BaseModel):

    tag_name: str = Field(
        ..., max_length=64, description="标签名称（在 object_type 内唯一）"
    )
    object_type: TagObjectType = Field(
        default=TagObjectType.SPACE, description='对象类型，当前仅支持 "space"'
    )
    object_id: str = Field(
        ..., max_length=255, description="被打标签对象ID（如 space_id）"
    )

    @field_validator("tag_name", mode="before")
    @classmethod
    def _normalize_tag_name(cls, v: Any) -> Any:
        if isinstance(v, str):
            s = v.strip().lower()
            if re.search(
                r"[~`!@#$%^&*()+={}\[\]|\\:;\"'<>,.?/￥…（）—【】「」‘’”“]", s
            ):
                raise ValueError("tag name 不允许包含空格或常见特殊字符")
            return s
        return v

    @field_validator("object_id", mode="before")
    @classmethod
    def _trim_object_id(cls, v: Any) -> Any:
        if isinstance(v, str):
            return v.strip()
        return v

    @model_validator(mode="before")
    def _normalize(cls, data: dict) -> dict:
        # 归一对象类型大小写，兼容字符串入参
        if isinstance(data.get("object_type"), str):
            data["object_type"] = data["object_type"].lower()
        return data


class Tagging(TimeStampedModel):
    tagging_id: str = Field(
        default_factory=lambda: str(uuid4()), description="标签绑定ID（UUID字符串）"
    )
    tenant_id: str = Field(..., max_length=64, description="所属租户ID")
    tag_id: str = Field(..., description="标签ID（FK -> tag.tag_id，UUID字符串）")
    tag_name: str = Field(
        ..., max_length=64, description="标签名称（在 object_type 内唯一）"
    )
    object_id: str = Field(
        ..., max_length=255, description="被打标签对象ID（如 space_id）"
    )
    object_type: TagObjectType = Field(
        default=TagObjectType.SPACE,
        description='对象类型，当前仅支持 "space"',
        max_length=32,
    )

    def update(self, **kwargs: Any) -> "Tagging":
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.updated_at = datetime.now(timezone.utc)
        return self

    @model_validator(mode="before")
    def _preprocess(cls, data: dict) -> dict:
        # UUID -> str
        for field, value in list(data.items()):
            if isinstance(value, UUID):
                data[field] = str(value)

        # 归一对象类型大小写
        if isinstance(data.get("object_type"), str):
            data["object_type"] = data["object_type"].lower()

        return data



import base64
import os
from typing import List, Union
from urllib.parse import quote

import chardet
import requests
from pydantic import HttpUrl

from whiskerrag_types.interface.loader_interface import BaseLoader
from whiskerrag_types.model import (
    GithubFileSourceConfig,
    Knowledge,
    KnowledgeSourceEnum,
)
from whiskerrag_types.model.knowledge import KnowledgeTypeEnum
from whiskerrag_types.model.knowledge_source import GithubRepoSourceConfig
from whiskerrag_types.model.multi_modal import Blob, Image, Text
from whiskerrag_utils.loader.git_repo_manager import get_repo_manager
from whiskerrag_utils.registry import RegisterTypeEnum, register


def file_to_text(full_path: str) -> Union[str, None]:
    try:
        with open(full_path, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError:
        pass
    try:
        with open(full_path, "rb") as f:
            content_bytes = f.read()
            guess = chardet.detect(content_bytes)
            encoding = guess["encoding"] if guess["confidence"] > 0.7 else None
            if encoding:
                try:
                    return content_bytes.decode(encoding)
                except Exception:
                    return None
    except Exception:
        return None
    return None


def file_to_base64(full_path: str) -> str:
    with open(full_path, "rb") as f:
        content = f.read()
        return base64.b64encode(content).decode("utf-8")


def get_github_default_branch(owner_repo: str) -> str:
    url = f"https://api.github.com/repos/{owner_repo}"
    headers = {"Accept": "application/vnd.github.v3+json"}
    resp = requests.get(url, headers=headers)
    if resp.ok:
        return resp.json()["default_branch"]  # type: ignore
    return "master"


def get_github_file_content(
    owner_repo: str,
    branch: str,
    path: str,
    knowledge_type: KnowledgeTypeEnum,
    token: Union[str, None] = None,
) -> Union[Text, Image, Blob]:
    url = f"https://api.github.com/repos/{owner_repo}/contents/{path}?ref={branch}"
    headers = {"Accept": "application/vnd.github.v3+json"}
    if token:
        headers["Authorization"] = f"Bearer {token}"
    resp = requests.get(url, headers=headers)
    if not resp.ok:
        raise RuntimeError(f"Github API获取文件失败: {resp.status_code} {resp.text}")
    info = resp.json()
    if info.get("encoding") == "base64" and "content" in info:
        content_b64 = info["content"]
        # github可能换行符分割base64，要去掉换行
        content_b64 = content_b64.replace("\n", "")
        content_bin = base64.b64decode(content_b64)
        if knowledge_type == KnowledgeTypeEnum.IMAGE:
            # Image类型兜底
            return Image(b64_json=content_b64, metadata={})
        else:
            # 尝试解码文本
            try:
                content_txt = content_bin.decode("utf-8")
                return Text(content=content_txt, metadata={})
            except Exception:
                guess = chardet.detect(content_bin)
                enc2 = guess["encoding"] if guess["confidence"] > 0.7 else None
                if enc2:
                    try:
                        content_txt = content_bin.decode(enc2)
                        return Text(content=content_txt, metadata={})
                    except Exception:
                        pass
                # Blob 兜底，推荐直接用 Blob.from_data
                # 带metadata/source/path都可以补上
                return Blob.from_data(content_bin, metadata={"path": path})
    else:
        # 极少数二进制 raw 返回
        raise RuntimeError("Github接口未返回base64编码内容，无法读取。")


def get_gitlab_file_content(
    url: str,
    owner_repo: str,
    branch: str,
    path: str,
    knowledge_type: KnowledgeTypeEnum,
    token: Union[str, None] = None,
) -> Union[Text, Image, Blob]:
    headers = {}
    if token:
        if isinstance(token, str) and token.startswith("git:"):
            token = token[4:]
        headers["PRIVATE-TOKEN"] = token
    # step1: 查项目id
    project_url = f"{url.rstrip('/')}/api/v3/project"
    resp = requests.get(project_url, params={"path": owner_repo}, headers=headers)
    if not resp.ok or "id" not in resp.json():
        raise RuntimeError(f"获取GitLab项目ID失败: {resp.status_code} {resp.text}")
    project_id = resp.json()["id"]
    if not branch:
        branch = resp.json().get("default_branch", "master")
    file_path_enc = quote(path, safe="")
    # step2: 查文件内容
    file_api_url = f"{url.rstrip('/')}/api/v4/projects/{project_id}/repository/files/{file_path_enc}"
    resp2 = requests.get(file_api_url, params={"ref": branch}, headers=headers)
    if not resp2.ok or "content" not in resp2.json():
        raise RuntimeError(f"获取GitLab文件失败: {resp2.status_code} {resp2.text}")
    info = resp2.json()
    content_b64 = info["content"]
    encoding = info["encoding"]
    content_b64 = content_b64.replace("\n", "")  # 防止意外有换行
    if encoding == "base64":
        content_bin = base64.b64decode(content_b64)
        if knowledge_type == KnowledgeTypeEnum.IMAGE:
            return Image(b64_json=content_b64, metadata={})
        else:
            try:
                content_txt = content_bin.decode("utf-8")
                return Text(content=content_txt, metadata={})
            except Exception:
                guess = chardet.detect(content_bin)
                enc2 = guess["encoding"] if guess["confidence"] > 0.7 else None
                if enc2:
                    try:
                        content_txt = content_bin.decode(enc2)
                        return Text(content=content_txt, metadata={})
                    except Exception:
                        pass
                return Blob.from_data(content_bin, metadata={"path": path})
    else:
        raise RuntimeError("暂不支持非base64编码内容")


@register(RegisterTypeEnum.KNOWLEDGE_LOADER, KnowledgeSourceEnum.GITHUB_FILE)
class GithubFileLoader(BaseLoader[Union[Text, Image, Blob]]):
    """
    从Git仓库加载单个文件的加载器
    """

    def __init__(self, knowledge: Knowledge):
        """
        初始化GithubFileLoader

        Args:
            knowledge: Knowledge实例，包含文件源配置

        Raises:
            ValueError: 无效的文件配置
        """
        if not isinstance(knowledge.source_config, GithubFileSourceConfig):
            raise ValueError("source_config should be GithubFileSourceConfig")
        self.knowledge = knowledge
        self.source_config = knowledge.source_config
        self.path = knowledge.source_config.path
        self.repo_name = knowledge.source_config.repo_name
        self.knowledge_type = knowledge.knowledge_type
        self.repo_manager = get_repo_manager()
        self.source_url = knowledge.source_config.url

        # 从文件配置中提取仓库配置
        self.repo_config = GithubRepoSourceConfig(
            **knowledge.source_config.model_dump(exclude={"path"}),
        )
        self.is_github = str(self.source_url).startswith("https://github.com")
        self.is_gitlab = not self.is_github

    async def load(self) -> List[Union[Text, Image, Blob]]:
        # 特殊处理图片
        if self.knowledge_type == KnowledgeTypeEnum.IMAGE and self.is_github:
            branch = self.source_config.branch
            repo_name = self.source_config.repo_name
            github_image_file_url = (
                f"https://raw.githubusercontent.com/{repo_name}/{branch}/{self.path}"
            )
            return [
                Image(
                    url=HttpUrl(github_image_file_url),
                    metadata=self.knowledge.metadata,
                )
            ]
        # 获取仓库本地路径
        repo_path = self.repo_manager.get_repo_path(self.repo_config)
        full_path = os.path.join(repo_path, self.path)
        # 1. 本地优先读
        if os.path.exists(full_path):
            text = file_to_text(full_path)
            if text is not None:
                return [Text(content=text, metadata=self.knowledge.metadata)]
            b64 = file_to_base64(full_path)
            if self.knowledge_type == KnowledgeTypeEnum.IMAGE:
                return [Image(b64_json=b64, metadata=self.knowledge.metadata)]
            else:
                return [Blob.from_data(b64, metadata=self.knowledge.metadata)]
        # 2. API 单文件获取兜底
        out = None
        if self.is_github:
            repo_name = self.source_config.repo_name
            branch = self.source_config.branch or get_github_default_branch(repo_name)
            # 直接用API获取单文件内容
            out = get_github_file_content(
                repo_name,
                branch,
                self.path,
                self.knowledge_type,
                self.source_config.auth_info,
            )
        elif self.is_gitlab:
            branch = self.source_config.branch or "master"
            out = get_gitlab_file_content(
                self.source_url,
                self.repo_name,
                branch,
                self.path,
                self.knowledge_type,
                self.source_config.auth_info,
            )
        else:
            raise ValueError("不支持的仓库类型，或请先本地clone仓库。")
        # 回填 metadata
        if out is not None:
            out.metadata = self.knowledge.metadata
            return [out]
        raise ValueError(f"File not found both local and remote: {self.path}")

    async def decompose(self) -> List[Knowledge]:
        """
        文件加载器不需要分解，返回空列表

        Returns:
            List[Knowledge]: 空列表
        """
        return []

    async def on_load_finished(self) -> None:
        """
        加载完成后的生命周期方法
        这里不清理仓库，因为可能还有其他文件需要访问
        """
        pass



import re
from datetime import datetime, timezone
from typing import Any, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, field_validator, model_validator

from whiskerrag_types.model.tagging import TagObjectType
from whiskerrag_types.model.timeStampedModel import TimeStampedModel


class TagCreate(BaseModel):
    name: str = Field(..., max_length=64, description="标签名称（租户内唯一）")
    description: Optional[str] = Field(
        default=None, max_length=255, description="标签描述"
    )
    object_type: TagObjectType = Field(
        default=TagObjectType.SPACE, description='标签作用对象类型（如 "space"）'
    )

    @field_validator("name", mode="before")
    @classmethod
    def _normalize_name(cls, v: Any) -> Any:
        if isinstance(v, str):
            s = v.strip().lower()
            # 禁止常见特殊字符
            if re.search(
                r"[~`!@#$%^&*()+={}\[\]|\\:;\"'<>,.?/￥…（）—【】「」‘’”“]", s
            ):
                raise ValueError("name 不允许包含空格或常见特殊字符")
            return s
        return v

    @model_validator(mode="before")
    def _normalize_object_type(cls, data: dict) -> dict:
        if isinstance(data.get("object_type"), str):
            data["object_type"] = data["object_type"].lower()
        return data


class Tag(TagCreate, TimeStampedModel):
    tag_id: str = Field(
        default_factory=lambda: str(uuid4()), description="标签ID（UUID字符串）"
    )
    tenant_id: str = Field(..., max_length=64, description="所属租户ID")
    object_type: TagObjectType = Field(
        default=TagObjectType.SPACE, description='标签作用对象类型（如 "space"）'
    )

    def update(self, **kwargs: Any) -> "Tag":
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.updated_at = datetime.now(timezone.utc)
        return self

    @model_validator(mode="before")
    def _convert_uuid_to_str(cls, data: dict) -> dict:
        for field, value in list(data.items()):
            if isinstance(value, UUID):
                data[field] = str(value)
        return data
